/********************************************************************************
 * Created on 2006 February 28
 *
 * Copyright (c) 2005 University of Southampton
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *******************************************************************************/
package org.pasoa.accessors;

import java.util.HashMap;
import java.util.Map;
import javax.xml.namespace.QName;
import org.pasoa.accessors.javacall.JavaCallXPathManager;
import org.pasoa.accessors.lfn.LFNAccessorManager;
import org.pasoa.accessors.snxpath.SingleNodeXPathManager;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * A registry for data accessor managers.  A data accessor is expected to
 * be encoded by an XML element with a given namespace and name.  These are
 * used to look up the data accessor manager for a given data accessor.
 *
 * For normal use, it is best to construct a DataAccessorRegistry by calling
 * createDefaultRegistry(), which sets up a registry with the default data
 * accessor managers (single node XPath and java call), rather than using the
 * constructor, which sets up an empty registry.
 *
 * A registry can have a default data accessor manager, which is used to
 * create data accessors for data items.
 */
public class DataAccessorRegistry implements DataAccessorManager {
    private Map                 _accessors;
    private DataAccessorManager _defaultManager;
    
    /**
     * Constructs an empty registry with no default data accessor manager.
     */
    public DataAccessorRegistry () {
        _accessors      = new HashMap ();
        _defaultManager = null;
    }
    
    private static DataAccessorRegistry _defaultRegistry = null;
    
    /** Constructs a registry with the two known data accessor managers ready-registered */
    public static DataAccessorRegistry getDefaultRegistry () {
        if (_defaultRegistry == null) {
            _defaultRegistry = new DataAccessorRegistry ();
            
            new SingleNodeXPathManager ().register (_defaultRegistry, true);
            new JavaCallXPathManager   ().register (_defaultRegistry, false);
            new LFNAccessorManager     ().register (_defaultRegistry, false);
        }
        
        return _defaultRegistry;
    }
    
    /** Converts the DOM element content of a data accessor to an XPath representation */
    public XPath convertToXPath (Element accessor) throws AccessorException {
        return lookup (accessor).convertToXPath (accessor);
    }
    
    /** Determines whether two data accessors (with given documentation styles) are equal */
    public boolean equalDataAccessors (Element accessor1, String style1, Element accessor2, String style2) throws AccessorException {
        if (accessor1 == null && accessor2 == null) {
            return true;
        }
        if (accessor1 == null || accessor2 == null) {
            return false;
        }
        return lookup (accessor1).equalDataAccessors (accessor1, style1, accessor2, style2);
    }
    
    /** Retrieves the data accessor for the data item in the content of a p-assertion */
    public Element getDataAccessorForItem (Node dataItem) throws AccessorException {
        if (_defaultManager == null) {
            throw new AccessorException ("No default data accessor manager set");
        } else {
            return _defaultManager.getDataAccessorForItem (dataItem);
        }
    }
    
    /**
     * Returns the default data accessor manager, or null if none specified.
     */
    public DataAccessorManager getDefault () {
        return _defaultManager;
    }
    
    /**
     * Retrieve the data accessor manager for a given data accessor element namespace-name.
     */
    public DataAccessorManager lookup (String namespace, String elementName) throws AccessorException {
        DataAccessorManager manager = lookup (new QName (namespace, elementName));
        
        if (manager == null) {
            throw new AccessorException ("Cannot find data accessor manager for qname {" + namespace + "}" + elementName);
        }
        
        return manager;
    }
    
    /**
     * Retrieve the data accessor manager for a given data accessor element
     */
    public DataAccessorManager lookup (Element accessorTag) throws AccessorException {
        if (accessorTag == null) {
            throw new AccessorException ("Cannot lookup data accessor manager for null data accessor");
        }
        return lookup (accessorTag.getNamespaceURI (), accessorTag.getLocalName ());
    }
    
    /**
     * Retrieve the data accessor manager for a given data accessor element namespace-name qname.
     */
    public DataAccessorManager lookup (QName accessorTagName) {
        return (DataAccessorManager) _accessors.get (accessorTagName);
    }
    
    /**
     * Register a data accessor for a given data accessor element namespace-name.
     */
    public void register (String namespace, String accessorTagName, DataAccessorManager manager, boolean useAsDefault) {
        register (new QName (namespace, accessorTagName), manager, useAsDefault);
    }
    
    /**
     * Register a data accessor for a given data accessor element namespace-name qname.
     */
    public void register (QName accessorTagName, DataAccessorManager manager, boolean useAsDefault) {
        _accessors.put (accessorTagName, manager);
        if (useAsDefault) {
            _defaultManager = manager;
        }
    }
}
