package org.pasoa.accessors;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * Represents an XPath, with two fields: a String for the path itself, and
 * a Map of prefixes to namespaces (String to String).
 */
public class XPath {
    private String _path;
    private Map    _namespaces;
    
    public XPath (String path, Map namespaces) {
        _path       = path;
        _namespaces = namespaces;
    }

    /**
     * Copy constructor.
     */
    public XPath (XPath original) {
        this (original.getPath (), new HashMap (original.getNamespaces ()));
    }
    
    /**
     * Generate a path that traverses the whole of the sub-tree described by
     * this element, using the given prefix-to-namespace mappings
     */
    public static XPath completeTraversal (Element root, HashMap spaces) {
        Map      rootSpace = new HashMap ();
        NodeList children  = root.getChildNodes ();
        XPath    path;
        Node     child;
        
        rootSpace.put ("ns1", root.getNamespaceURI ());
        path = new XPath ("/ns1:" + root.getLocalName (), rootSpace);
        path = removePrefixes (path, spaces.keySet ());
        
        for (int index = 0; index < children.getLength (); index += 1) {
            child = children.item (index);
            if (child instanceof Text) {
                path = new XPath (path.getPath () + "[text()=\"" + child.getNodeValue () + "\"]", path.getNamespaces ());
            }
            if (child instanceof Element) {
                path = merge (path, completeTraversal ((Element) child, spaces), true);
            }
        }
        
        return path;
    }
    
    public Map getNamespaces () {
        return _namespaces;
    }
    
    public String getPath () {
        return _path;
    }

    /**
     * Merges a search for a node (head) with a search within that node (tail).
     * This can either be expressed as a test on head (asTest==true) or as
     * a traversal to the end of the tail (asTest=false)
     */
    public static XPath merge (XPath head, XPath tail, boolean asTest) {
        XPath newHead  = removePrefixes (head, tail.getNamespaces ().keySet ());
        Map   mergedNS = new HashMap (tail.getNamespaces ());
        
        mergedNS.putAll (newHead.getNamespaces ());
        
        if (asTest) {
            return new XPath (newHead.getPath () + "[" + tail.getPath ().substring (1) + "]", mergedNS);
        } else {
            return new XPath (newHead.getPath () + tail.getPath (), mergedNS);
        }
    }
    
    public static String newPrefix (Collection existing) {
        int index = 1;
        
        while (existing.contains ("ns" + index)) {
            index += 1;
        }
        
        return "ns" + index;
    }
    
    private static XPath removePrefixes (XPath original, Collection bannedPrefixes) {
        Collection allPrefixes = new HashSet (original.getNamespaces ().keySet ());
        String     newPath     = original.getPath ();
        Map        newSpaces   = new HashMap (original.getNamespaces ());
        Iterator   eachKey     = bannedPrefixes.iterator ();
        Object     key;
        String     newKey;
        
        allPrefixes.addAll (bannedPrefixes);
        while (eachKey.hasNext ()) {
            key = eachKey.next ();
            if (newSpaces.containsKey (key)) {
                newKey = newPrefix (allPrefixes);
                newPath.  replaceAll (key + ":", newKey);
                newSpaces.put (newKey, original.getNamespaces ().get (key));
                newSpaces.remove (key);
            }
        }
        
        return new XPath (newPath, newSpaces);
    }
    
    public String toString () {
        return "Path: " + _path + " with namespaces " + _namespaces;
    }
}
