/********************************************************************************
 * Created on 2006 February 28
 *
 * Copyright (c) 2005 University of Southampton
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *******************************************************************************/
package org.pasoa.accessors.snxpath;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import javax.xml.namespace.QName;
import org.pasoa.common.Constants;
import org.pasoa.accessors.DataAccessorManager;
import org.pasoa.accessors.DataAccessorRegistry;
import org.pasoa.accessors.XPath;
import org.pasoa.simpledom.DOMDeserialiser;
import org.pasoa.simpledom.DOMSerialiser;
import org.pasoa.simpledom.DOMTraverser;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

public class SingleNodeXPathManager implements DataAccessorManager {
    public static final String _namespace   = Constants.PQUERY_NS;
    public static final String _elementName = "singleNodeXPath";
    
    public Element createAccessor (XPath xpath) {
        return createAccessor (xpath.getPath (), xpath.getNamespaces ());
    }
    
    public Element createAccessor (String path, Map namespaces) {
        DOMSerialiser writer     = new DOMSerialiser ();
        Iterator      eachPrefix = namespaces.keySet ().iterator ();
        String        prefix;
        
        writer.namespace (_namespace);
        writer.start (_elementName);
        writer.single ("path", path);
        while (eachPrefix.hasNext ()) {
            prefix = (String) eachPrefix.next ();
            writer.start ("namespaceMapping");
            writer.single ("prefix",    prefix);
            writer.single ("namespace", (String) namespaces.get (prefix));
            writer.end ();
        }
        
        return ((Document) writer.getRoot ()).getDocumentElement ();
    }
    
    public XPath convertToXPath (Element accessor) {
        if (accessor == null) {
            return null;
        }
        
        DOMDeserialiser parse  = new DOMDeserialiser (_namespace);
        String          path   = parse.text (accessor, "path");
        Iterator        eachNS = parse.iterator (accessor, "namespaceMapping");
        Map             spaces = new HashMap ();
        Element         map;
        
        while (eachNS.hasNext ()) {
            map = (Element) eachNS.next ();
            spaces.put (parse.text (map, "prefix"), parse.text (map, "namespace"));
        }
        
        return new XPath (path, spaces);
    }
    
    private String expandedForm (XPath xpath) {
        if (xpath == null) {
            return "null";
        }
        
        String expanded = xpath.getPath ();
        int    start    = 0;
        int    colon;
        int    slash;
        String prefix;
        String namespace;
        
        while (start >= 0 && expanded.indexOf (":", start) >= 0) {
            colon     = expanded.indexOf (":", start);
            slash     = expanded.lastIndexOf ("/", colon);
            prefix    = expanded.substring (slash + 1, colon);
            namespace = (String) xpath.getNamespaces ().get (prefix);
            if (namespace == null) {
                expanded  = expanded.substring (0, slash + 1) + " " + expanded.substring (colon + 1);
                start     = expanded.indexOf ("/", slash + 1);
                //throw new RuntimeException ("SingleNodeXPath: No namespace for prefix " + prefix + " in path " + xpath.getPath ());
            } else {
                expanded  = expanded.substring (0, slash + 1) + " " + namespace + " " + expanded.substring (colon + 1);
                start     = expanded.indexOf ("/", slash + namespace.length () + 3);
            }
        }
        
        return expanded;
    }
    
    public boolean equalDataAccessors (Element accessor1, String style1, Element accessor2, String style2) {
        XPath first  = convertToXPath (accessor1);
        XPath second = convertToXPath (accessor2);
        
        return (first == null && second == null) ||
                (first != null && second != null && expandedForm (first).equals (expandedForm (second)));
    }
    
    /**
     * Generate a path leading to node (inclusive) from root (exclusive)
     */
    private static XPath generate (Node node, Node root) {
        return generate (node, root, new HashMap ());
    }
    
    /**
     * Generate a path leading to node (inclusive) from root (exclusive) using
     * the given prefix to namespace mappings (new spaces may be added to this
     * mapping)
     */
    private static XPath generate (Node node, Node root, Map namespaces) {
        Node     parent     = node.getParentNode ();
        int      index      = 0;
        int      count      = 0;
        String   prefix     = null;
        Iterator eachPrefix;
        XPath    part;
        boolean  found;
        
        if (parent == null) {
            return new XPath (null, new HashMap ());
        }
        if (!(root != null && node.isSameNode (root)) && parent != null && !(parent instanceof Document)) {
            index = DOMTraverser.indexOfChild (parent, node);
        }
        
        if (node instanceof Element) {
            eachPrefix = namespaces.keySet ().iterator ();
            found      = false;
            while (eachPrefix.hasNext ()) {
                prefix = (String) eachPrefix.next ();
                if ((namespaces.get (prefix) == null || namespaces.get (prefix).equals (""))) {
                    if ((node.getNamespaceURI () == null || node.getNamespaceURI ().equals (""))) {
                        found = true;
                        break;
                    }
                } else {
                    if (namespaces.get (prefix).equals (node.getNamespaceURI ())) {
                        found = true;
                        break;
                    }
                }
            }
            if (!found) {
                prefix = XPath.newPrefix (namespaces.keySet ());
            }
            namespaces.put (prefix, node.getNamespaceURI ());
            part = new XPath ("/" + prefix + ":" + node.getLocalName () + "[0]", namespaces);
        } else {
            part = new XPath ("/text()[" + index + "]", namespaces);
        }
        
        if (parent != null && parent.isSameNode (root)) {
            return part;
        }
        if (!(root != null && node.isSameNode (root)) && parent != null && !(parent instanceof Document)) {
            return XPath.merge (generate (parent, root, namespaces), part, false);
        } else {
            return part;
        }
    }
    
    public Element getDataAccessorForItem (Node dataItem) {
        Node root = DOMTraverser.getAncestorNode (dataItem, Constants.PSTRUCT_NS, new String[] {"content", "relationshipPAssertion"});
        
        if (root != null) {
            return createAccessor (generate (dataItem, root));
        } else {
            return null;
        }
    }
    
    public void register (DataAccessorRegistry registry, boolean useAsDefault) {
        registry.register (new QName (_namespace, _elementName), this, useAsDefault);
    }
}
