package org.pasoa.preserv.pquery;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import org.pasoa.accessors.AccessorException;
import org.pasoa.accessors.DataAccessorRegistry;
import org.pasoa.preserv.index.IndexProblem;
import org.pasoa.preserv.index.RelationshipIndex;
import org.pasoa.pstructure.PAssertionDataKey;
import org.pasoa.pstructure.RelationshipPAssertion;
import org.pasoa.storeclient.proxy.StoreCatalogue;

public class ProvenanceQueryEngine implements ProvenanceQuery {
    private StoreCatalogue       _stores;
    private RelationshipIndex    _relationships;
    private String               _thisStore;
    private DataAccessorRegistry _accessors;
    private boolean              _locallyDriven;
    
    public ProvenanceQueryEngine (StoreCatalogue stores, String thisStore, boolean locallyDriven) {
        _stores        = stores;
        _accessors     = DataAccessorRegistry.getDefaultRegistry ();
        _thisStore     = thisStore;
        _relationships = new RelationshipIndex (stores, _accessors);
        _locallyDriven = locallyDriven;
    }
    
    /**
     * @throws PQueryProblem if an element of the provenance query request is malformed
     */
    public ProvenanceQueryResults resolve (ProvenanceQueryRequest request) throws PQueryProblem {
        QueryDataHandle        qdh     = request.getQueryDataHandle ();
        ProvenanceQueryResults results = resolve (qdh.resolve (_accessors),
                request.getRelationshipTargetFilter (), new HashSet ());
        
        System.out.println ("Query Data Handle: " + qdh.getXPath ());
        System.out.println ("Relationship Target Filter: " + request.getRelationshipTargetFilter ().getXPath ());
        System.out.println ("Starts:        " + results.getStarts ().size ());
        System.out.println ("Relationships: " + results.getFullRelationships ().size ());
        
        return results;
    }
    
    /**
     * @throws PQueryProblem if relationship target filter is malformed
     */
    private ProvenanceQueryResults resolve (Collection starts,
            RelationshipTargetFilter scope, Collection excludes) throws PQueryProblem {
        Collection         results   = new HashSet ();
        Iterator           eachStart = starts.iterator ();
        Object             item;
        PAssertionDataKey  padk;
        
        while (eachStart.hasNext ()) {
            try {
                results.addAll (resolve ((PAssertionDataKeyTemplate) eachStart.next (), _thisStore,
                        scope, excludes));
            } catch (IndexProblem ex) {
                System.err.println ("Could not find provenance of a data item discovered by query data handle");
                ex.printStackTrace ();
            }
        }
        
        return new ProvenanceQueryResults (starts, results);
    }
    
    /**
     * Find the provenance (full relationships) of a given p-assertion data item
     * at a given location.
     *
     * @throws PQueryProblem if the provenance query locally or remotely fails
     * @throws IndexProblem  if cannot retrieve relationships of the start item
     */
    private Collection resolve (Object start, String location,
            RelationshipTargetFilter scope, Collection exclude) throws IndexProblem, PQueryProblem {
        if (_locallyDriven || location.equals (_thisStore)) {
            return resolveLocal (start, location, scope, exclude);
        } else {
            return _stores.getStoreProxy (location).getPQueryInterface ().resolve (
                    new ProvenanceQueryRequest (new QueryDataHandle ((PAssertionDataKey) start), scope)).getFullRelationships ();
        }
    }
    
    /**
     * Find the provenance (full relationships) of a given p-assertion data item
     * at a given location, with the querying being driven from this process.
     * This will normally only happen for p-assertion data items that are contained
     * in this provenance store, but depends on the engine configuration.
     *
     * @throws PQueryProblem if relationship target filter is malformed
     * @throws IndexProblem  if cannot retrieve relationships of the start item
     */
    private Collection resolveLocal (Object start, String location,
            RelationshipTargetFilter scope, Collection exclude) throws IndexProblem, PQueryProblem {
        Collection        results      = new HashSet ();
        Iterator          eachRelation = _relationships.getRelationships (location, start).iterator ();
        PAssertionDataKey padk;
        
        while (eachRelation.hasNext ()) {
            if (start instanceof PAssertionDataKey) {
                padk = (PAssertionDataKey) start;
            } else {
                try {
                    padk = ((PAssertionDataKeyTemplate) start).getInstantiation ();
                } catch (AccessorException accessorProblem) {
                    throw new PQueryProblem ("Problem resolving data accessor for start", accessorProblem);
                }
            }
            results.addAll (resolve (padk, location, (RelationshipPAssertion) eachRelation.next (), scope, exclude));
        }
        
        return results;
    }
    
    /**
     * @throws PQueryProblem if relationship target filter is malformed
     * @throws IndexProblem  if cannot retrieve relationships targets of relationship
     */
    private Collection resolve (PAssertionDataKey effectKey, String effectLocation,
            RelationshipPAssertion relationship,
            RelationshipTargetFilter scope, Collection exclude) throws IndexProblem, PQueryProblem {
        Collection targets    = _relationships.getRelationshipTargets (relationship, effectLocation);
        Iterator   eachTarget = targets.iterator ();
        Collection results    = new HashSet ();
        
        while (eachTarget.hasNext ()) {
            try {
                results.addAll (resolve (effectKey, effectLocation,
                        (RelationshipTarget) eachTarget.next (), relationship, scope, exclude));
            } catch (MalformedURLException ex) {
                System.err.println ("Link of relationship target was malformed");
                ex.printStackTrace ();
            } catch (IndexProblem ex) {
                System.err.println ("Could not iteratively find relationships for a target");
                ex.printStackTrace ();
            }
        }
        
        return results;
    }
    
    /**
     * @throws PQueryProblem         if relationship target filter is malformed
     * @throws MalformedURLException if target link URL is malformed
     * @throws IndexProblem          if cannot retrieve relationships (so find provenance) of target
     */
    private Collection resolve (PAssertionDataKey effect, String effectLocation,
            RelationshipTarget target, RelationshipPAssertion relationship,
            RelationshipTargetFilter scope, Collection excludes) throws PQueryProblem, MalformedURLException, IndexProblem {
        Collection results = new HashSet ();
        
        if (scope.accept (target)) {
            results.add (new FullRelationship (effect, effectLocation.toString (),
                    relationship, target));
            if (!excludes.contains (target)) {
                excludes.add (target);
                results.addAll (resolve (target, target.getLink ().getStoreURL (), scope, excludes));
            }
        }
        
        return results;
    }
}
