/********************************************************************************
 * Created on Jul 6, 2005
 *
 * Copyright (c) 2005 University of Southampton
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to deal 
 * in the Software without restriction, including without limitation the rights 
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell 
 * copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS 
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
 * SOFTWARE.
 *******************************************************************************/
package org.pasoa.preserv.processor;

import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.StringReader;
import java.util.Vector;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.pasoa.preserv.factory.PlugIn;
import org.pasoa.preserv.record.CacheShutdownException;
import org.pasoa.preserv.record.CacheableItem;
import org.pasoa.preserv.storage.GeneralStorageSystem;
import org.pasoa.preserv.storage.StorageSystem;
import org.pasoa.simpledom.DOMWriter;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;

/**
 * 
 * @author Paul Groth pg03r@ecs.soton.ac.uk
 */
public class Dispatcher implements Runnable
{
    private long process_cache_items_wait_time = 75;
    private boolean _processCachedItems;
    private boolean _cachedItemsOpen;
    private PlugInRegistry _pir;
    private StorageSystem _ss;

    /*
     * The DocumentBuilder for the processAndDispatch method
     */
    private DocumentBuilder _docBuilder;

    /*
     * The DocumentBuilder for the dispatcher thread that handles the cache.
     * Seperate DocumentBuilders are necessary because implemenatations of the
     * abstract class are not guaranteed to be thread safe.
     */
    private DocumentBuilder _threadDocBuilder;
    private Vector _responses;
    private Thread _thisThread;
    private DocumentBuilderFactory _dbf;
    
    public Dispatcher(StorageSystem ss, PlugInRegistry pir) {
        _ss = ss;
        _pir = pir;
        _responses = new Vector(100);
        _processCachedItems = true;
        _dbf = DocumentBuilderFactory.newInstance();
        
        // We have to let the underlying parser know that it should be namespace
        // aware.
        _dbf.setNamespaceAware(true);

        try {
            _threadDocBuilder = _dbf.newDocumentBuilder();
        } catch (Exception e) {
            System.out.println("PReServ:Dispatcher: Unable to create document builders");
            e.printStackTrace();
        }

        // Start the Dispatch Thread
        try {
             _thisThread = new Thread(this, "PReServ Dispatch Thread");
             _thisThread.setPriority(Thread.MIN_PRIORITY);
             _thisThread.start();
        } catch (Exception threadExp) {
            System.out.println("PReServ:Dispatcher: Unable to start thread");
            threadExp.printStackTrace();
        }
    }


    public void processAndDispatch(InputStream is, 
                                   OutputStream os,
                                   String context) throws Exception 
    {
        processAndDispatch (is, os, context, false);   
    }
    
    /**
     * Processes a SOAP message from the given input stream and dispatches the
     * contents to the appropriate PlugIn specified by the context.
     * 
     * @param is
     * @throws Exception
     */
    public void processAndDispatch(InputStream is, 
                                   OutputStream os,
                                   String context,
                                   boolean compressed) throws Exception 
    {
        
        
        PlugIn pi = _pir.lookup(context);

        if (pi.receiveDispatchesSyncronously()) {
            Document doc = _dbf.newDocumentBuilder().parse(is);
            Document responseDoc = pi.process(doc, _ss);
            // System.out.println (DOMWriter.writeToString(responseDoc));
            DOMWriter domWriter = new DOMWriter(os);
            // domWriter.setFormatted(true);
            domWriter.write(responseDoc);

            // _docBuilder.reset();
        } else {
           /* BufferedReader br = new BufferedReader(new InputStreamReader(is));
            StringBuffer sb = new StringBuffer();
            String line;

            while ((line = br.readLine()) != null) {
              sb.append(line);
            }
            */
            
            //System.out.println ("Caching: " + sb.toString());
            //CacheableItem ci = new CacheableItem(context, sb);
            //long startTime = System.currentTimeMillis();
             //_ss.cache(ci);
            _ss.cache(is, context, compressed);
            
           //long endTime = System.currentTimeMillis();
            //System.out.println("Cache time: " + (endTime - startTime));
            PrintWriter pw = new PrintWriter(os);
            pw.print(pi.getDefaultResponse());
            pw.flush();
        }
       
    }

    public void run() {
        CacheableItem ci;
        while (_processCachedItems) {
            try {
                ci = _ss.getCachedItem();
                processCacheableItem (ci);
                Thread.yield();
            } 
            catch (CacheShutdownException cse) {
                System.out.println("PReServ:Dispatcher: Thread was blocked, and cache shutdown excpetion was thrown now exiting");
                return;    
            }
            catch (Exception e) {
                System.out.println("PReServ:Dispatcher: Problems in Dispatch Thread");
                e.printStackTrace();
            }
        }
       
    }

    private void processCacheableItem (CacheableItem ci)
       throws Exception
    {
        String docStr = ci.getItem().toString().trim ();
        //System.out.println (docStr);
        Document doc = _threadDocBuilder.parse(new InputSource(
                  new StringReader(docStr)));
        PlugIn pi = _pir.lookup(ci.getContext());
        Document responseDoc = pi.process(doc, _ss);
        // _responses.addElement(responseDoc);
        // _threadDocBuilder.reset();
    }
    
    public Document getResponse() {
        if (_responses.isEmpty()) {
            return null;
        } else {
            Document doc = (Document) _responses.firstElement();
            return doc;
        }
    }

    public int getNumberOfResponses() {
        return _responses.size();
    }

    public void shutdown() throws Exception {
        _processCachedItems = false;
        _ss.cacheShutdown();
        _thisThread.join();
        // remember to call shutdown on the storage system as well
    }

    public long getProcessCacheItemsWaitTime() {
        return process_cache_items_wait_time;
    }

    public void setProcessCacheItemsWaitTime(long waitTime) {
        process_cache_items_wait_time = waitTime;
    }
}
