/********************************************************************************
 * Created on 2006 February 28
 *
 * Copyright (c) 2005 University of Southampton
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *******************************************************************************/
package org.pasoa.preserv.test;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.UnsupportedEncodingException;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import org.pasoa.pstructure.Record;
import org.pasoa.preserv.processor.Dispatcher;
import org.pasoa.preserv.processor.PlugInRegistry;
import org.pasoa.preserv.pquery.ProvenanceQueryDeserialiser;
import org.pasoa.preserv.pquery.ProvenanceQueryPlugin;
import org.pasoa.preserv.pquery.ProvenanceQueryRequest;
import org.pasoa.preserv.xquery.XQueryPlugIn;
import org.pasoa.preserv.record.RecordPlugIn;
import org.pasoa.preserv.storage.GeneralStorageSystem;
import org.pasoa.preserv.storage.StorageSystem;
import org.pasoa.preserv.storage.je.BerkeleyDBJEStorage2;
import org.pasoa.pstructure.dom.PStructureJavaModel;
import org.pasoa.util.httpsoap.DefaultDocumentClient;
import org.pasoa.storeclient.messages.ProvenanceQueryRequestDocument;
import org.pasoa.storeclient.messages.RecordRequestDocument;
import org.pasoa.storeclient.messages.XPathRequestDocument;
import org.pasoa.storeclient.messages.XQueryRequestDocument;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * A testbed implementation that uses an in-memory provenance store
 * backed by a database stored in a newly created temporary directory.
 */
public class LocalTestbed extends Testbed {
    public static final String _localURL = "http://in.memory";
    
    private Dispatcher            _dispatch;
    private StorageSystem         _store;
    private List <File>           _temps;
    private String                _home;
    private RecordPlugIn          _record;
    private BerkeleyDBJEStorage2  _backend;
    
    public LocalTestbed () throws Exception {
        this (true);
    }
    
    public LocalTestbed (boolean compress) throws Exception {
        String tmpPath = this.createTemporaryDirectory ().getAbsolutePath ();
        intialize (tmpPath, false, compress);
    }
    
    public LocalTestbed (String path, boolean compress) throws Exception {
        intialize (path, false, compress);
    }
    
    public LocalTestbed (String path) throws Exception {
        this (path, true);
    }
    
    public LocalTestbed (boolean disableVerifiedCommits, boolean compress) throws Exception {
        String tmpPath = this.createTemporaryDirectory ().getAbsolutePath ();
        intialize (tmpPath, disableVerifiedCommits, compress);
        
    }
    
    public LocalTestbed (String path, boolean disableVerifiedCommits, boolean compress) throws Exception {
        intialize (path, disableVerifiedCommits, compress);
    }
    
    private void intialize (String path, boolean disableVerifiedCommits, boolean compress) throws Exception {
        _home = path;
        _backend = new BerkeleyDBJEStorage2 (_home, disableVerifiedCommits, compress);
        _record  = new RecordPlugIn ();
        ProvenanceQueryPlugin  query    = new ProvenanceQueryPlugin ();
        XQueryPlugIn           xquery   = new XQueryPlugIn ();
        PlugInRegistry         plugins  = new PlugInRegistry ();
        
        plugins.registerPlugIn ("record", _record);
        plugins.registerPlugIn ("pquery", query);
        plugins.registerPlugIn ("xquery", xquery);
        
        _store    = new GeneralStorageSystem (_backend, "org.pasoa.preserv.xquery.laxquery.XQueryEngine");
        _dispatch = new Dispatcher (_store, plugins);
    }
    
    public String getPathToStorageDirectory () {
        return _home;
    }
    
    public String getURL () {
        return _localURL;
    }
    
    public void setSynchronousRecording (boolean sync) {
        _record.setReceiveDipatchesSynchronously (sync);
    }
    
    public void enableBackendRecordTimeLogging (PrintWriter output) {
        _backend.enableRecordTimeLog (output);
    }
    
    /**
     * Deletes temporary files and directories created using createTemporaryDirectory,
     * including the one containing the provenance store database.
     */
    public void cleanUp () {
        try {
            _dispatch.shutdown ();
            _store.shutdown ();
        } catch (Exception cannotShutdown) {
            System.err.println ("Cannot shutdown " + cannotShutdown.getMessage ());
        }
    }
    
    /**
     * Creates a temporary directory.  Used by LocalTestbed to create a directory
     * in which to keep the provenance store database, but can also be used to
     * create directories for other data.
     */
    public File createTemporaryDirectory () throws IOException {
        File temp = File.createTempFile ("database", "");
        temp.delete ();
        temp.mkdir ();
        temp.deleteOnExit ();
        
        if (_temps == null) {
            _temps = new LinkedList ();
        }
        _temps.add (temp);
        
        return temp;
    }
    
    /**
     * Convenience method to return the whole provenance store contents as a p-structure DOM document.
     */
    public Document getStoreContents () throws Exception {
        return PStructureJavaModel.getPStructureDocument (_store.getRetrievalInterface ());
    }
    
    public Element provenanceQuery (String qdhXPath, String scopeXPath, Map namespaces) throws UnsupportedEncodingException, Exception {
        ProvenanceQueryDeserialiser    parser     = new ProvenanceQueryDeserialiser ();
        ProvenanceQueryRequestDocument request;
        Document                       response;
        
        request  = new ProvenanceQueryRequestDocument (qdhXPath, scopeXPath, namespaces);
        response = new DefaultDocumentClient ().invoke (_dispatch, "pquery", request);
        
        return response.getDocumentElement ();
        //return parser.deserialiseProvenanceQueryResults (response.getDocumentElement ());
    }

    public Element provenanceQuery (ProvenanceQueryRequest query) throws UnsupportedEncodingException, Exception {
        ProvenanceQueryDeserialiser    parser     = new ProvenanceQueryDeserialiser ();
        ProvenanceQueryRequestDocument request;
        Document                       response;
        
        request  = new ProvenanceQueryRequestDocument (query);
        response = new DefaultDocumentClient ().invoke (_dispatch, "pquery", request);
        
        return response.getDocumentElement ();
        //return parser.deserialiseProvenanceQueryResults (response.getDocumentElement ());
    }
    
    public void record (Record record) throws UnsupportedEncodingException, Exception {
        RecordRequestDocument request = new RecordRequestDocument (record);
//System.out.println (DOMWriter.writeToString (request, true));
        new DefaultDocumentClient ().invoke (_dispatch, "record", request);
    }
    
    public Element xpath (String xpath, String root, String item, Map namespaces) throws UnsupportedEncodingException, Exception {
        XPathRequestDocument request = new XPathRequestDocument (xpath, root, item, namespaces);
        
        return new DefaultDocumentClient ().invoke (_dispatch, "xquery", request).getDocumentElement ();
    }
    
    public Element xquery (String xquery) throws UnsupportedEncodingException, Exception {
        XQueryRequestDocument request = new XQueryRequestDocument (xquery);
        
        return new DefaultDocumentClient ().invoke (_dispatch, "xquery", request).getDocumentElement ();
    }
}
