/********************************************************************************
 * Created on 2006 February 28
 *
 * Copyright (c) 2005 University of Southampton
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *******************************************************************************/
package org.pasoa.preserv.test;

import java.io.IOException;
import java.io.StringReader;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.pasoa.accessors.snxpath.SingleNodeXPathManager;
import org.pasoa.common.Constants;
import org.pasoa.preserv.pquery.ProvenanceQueryRequest;
import org.pasoa.pstructure.ActorStatePAssertion;
import org.pasoa.pstructure.ExposedInteractionMetadata;
import org.pasoa.pstructure.GlobalPAssertionKey;
import org.pasoa.pstructure.InteractionKey;
import org.pasoa.pstructure.InteractionPAssertion;
import org.pasoa.pstructure.Link;
import org.pasoa.pstructure.ObjectID;
import org.pasoa.pstructure.PHeader;
import org.pasoa.pstructure.Record;
import org.pasoa.pstructure.RecordExposedInteractionMetadata;
import org.pasoa.pstructure.RelationshipPAssertion;
import org.pasoa.pstructure.SubjectID;
import org.pasoa.pstructure.serialiser.PStructureSerialiser;
import org.pasoa.storeclient.messages.SOAPMessageWithPHeader;
//import org.pasoa.accessors.snxpath.SingleNodeXPath;
import org.pasoa.util.httpsoap.WSAddressEndpoint;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Provides simple methods for writing a simulator for p-assertion recording,
 * using particular limiting assumptions.
 * Also provides methods to do XQueries, XPaths and provenance queries.
 * This class cannot be used directly: use LocalTestbed for a local provenance
 * store (not deployed as a Web Service) or RemoteTestbed for a Web Service
 * provenance store.
 */
public abstract class Testbed {
    private int                 _interaction;
    private int                 _lpaid;
    private DocumentBuilder     _domBuilder;
    
    public static final String verbatimStyle = "http://www.pasoa.org/documentationStyles/verbatim";
    public static final String allContents   = "$ps:pstruct";
    
    public Testbed () throws Exception {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance ();
        
        _interaction = 0;
        _lpaid       = 100;
        factory.setNamespaceAware (true);
        _domBuilder  = factory.newDocumentBuilder ();
    }
    
    private List addObjectIDToList (List objectIDs, GlobalPAssertionKey gpak, String parameter, Element accessor, String link) {
        if (link == null || link.equals (LocalTestbed._localURL)) {
            objectIDs.add (new ObjectID (gpak, parameter, accessor, null));
        } else {
            objectIDs.add (new ObjectID (gpak, parameter, accessor, new Link (link)));
        }
        return objectIDs;
    }
    
    public Element createDataAccessor (String path, Map spaces) {
        return new SingleNodeXPathManager ().createAccessor (path, spaces);
    }
    
    /**
     * Creates an XML data accessor based on an XPath, for use in relationship p-assertions.
     */
    public Element createDataAccessor (String path, String prefix, String namespace) {
        Map spaces = new HashMap ();
        
        spaces.put (prefix, namespace);
        spaces.put ("ps", Constants.PSTRUCT_NS);
        
        return createDataAccessor (path, spaces);
    }
    
    public abstract String getURL ();
    
    private String getViewKind (boolean isSender) {
        if (isSender) {
            return Constants.SENDER_VIEW_TYPE;
        } else {
            return Constants.RECEIVER_VIEW_TYPE;
        }
    }
    
    /**
     * Creates a new interaction ID.
     */
    protected String nextInteractionID () {
        _interaction += 1;
        return Integer.toString (_interaction);
    }
    
    /**
     * Creates a new local p-assertion ID.
     */
    protected String nextLocalPAssertionID () {
        _lpaid += 1;
        return Integer.toString (_lpaid);
    }
    
    /**
     * Performs a provenance query, as per the behaviour described in ClientLib's Javadoc.
     */
    public abstract Element provenanceQuery (String qdhXPath, String scopeXPath, Map namespaces) throws Exception;
    
    public abstract Element provenanceQuery (ProvenanceQueryRequest query) throws Exception;
    
    /**
     * Records a p-assertion to the provenance store, as per the behaviour described in ClientLib's Javadoc.
     */
    public abstract void record (Record record) throws Exception;
    
    public Record recordActorState (Record interaction, String xmlState) throws Exception {
        ActorStatePAssertion passertion = new ActorStatePAssertion (nextLocalPAssertionID (), toDOM (xmlState));
        Record               record     = new Record (passertion, interaction.getInteractionKey (), interaction.getViewKind (), interaction.getAsserter ());
        
        record (record);
        
        return record;
    }
    
    /**
     * Records two interaction p-assertions regarding the message (XML encoded as a string),
     * one from the
     * sender's view, one from the recorder's view.  A WS-Addressing endpoint
     * containing an actor's URL is used for both message source/sink and asserter
     * ID.  The documentation style is verbatim.
     */
    public RecordViews recordInteraction (String senderURL, String receiverURL, String xmlMessage) throws Exception {
        return recordInteraction (senderURL, receiverURL, toDOM (xmlMessage));
    }
    
    /**
     * Record two views of an interaction in this store.
     */
    public RecordViews recordInteraction (String senderURL, String receiverURL, Element messageBody) throws Exception {
        return recordInteraction (senderURL, this, receiverURL, this, messageBody);
    }
    
    public RecordViews recordInteraction (String senderURL, Testbed senderStore,
            String receiverURL, Testbed receiverStore, String messageBody) throws Exception {
        return recordInteraction (senderURL, senderStore, receiverURL, receiverStore, toDOM (messageBody));
    }
    
    /**
     * Record two views of an interaction in two provenance stores, with view links
     * between them and expose the metadata.  Note that, unlike other recordInteraction methods, the message
     * body, rather than the whole message, is given.
     */
    public RecordViews recordInteraction (String senderURL, Testbed senderStore,
            String receiverURL, Testbed receiverStore, Element messageBody) throws Exception {
        // Construct interaction metadata (view links)
        List                       senderMetadata   = new LinkedList ();
        List                       receiverMetadata = new LinkedList ();
        PStructureSerialiser       writer           = new PStructureSerialiser ();
        String                     senderPSURL      = senderStore.getURL ();
        String                     receiverPSURL    = receiverStore.getURL ();
        Element                    senderLink       = writer.serialiseViewLink (new Link (senderPSURL)).getDocumentElement ();
        Element                    receiverLink     = writer.serialiseViewLink (new Link (receiverPSURL)).getDocumentElement ();

        senderMetadata.  add (receiverLink);
        receiverMetadata.add (senderLink);
        
        // Construct messages containing p-headers
        WSAddressEndpoint          sender           = new WSAddressEndpoint (senderURL);
        WSAddressEndpoint          receiver         = new WSAddressEndpoint (receiverURL);
        InteractionKey             key              = new InteractionKey (sender.getElement (), receiver.getElement (), nextInteractionID ());
        PHeader                    senderPH         = new PHeader (key, senderMetadata,   new LinkedList ());
        PHeader                    receiverPH       = new PHeader (key, receiverMetadata, new LinkedList ());
        String                     lpaid            = nextLocalPAssertionID ();
        Element                    senderMessage    = new SOAPMessageWithPHeader (messageBody, senderPH).getDocumentElement ();
        Element                    receiverMessage  = new SOAPMessageWithPHeader (messageBody, receiverPH).getDocumentElement ();
        InteractionPAssertion      senderIPA        = new InteractionPAssertion (lpaid, verbatimStyle, senderMessage);
        InteractionPAssertion      receiverIPA      = new InteractionPAssertion (lpaid, verbatimStyle, receiverMessage);
        Element                    senderAsserter   = sender.getElement ();
        Element                    receiverAsserter = receiver.getElement ();
        Record                     senderRecord     = new Record (senderIPA,   key, Constants.SENDER_VIEW_TYPE,   senderAsserter);
        Record                     receiverRecord   = new Record (receiverIPA, key, Constants.RECEIVER_VIEW_TYPE, receiverAsserter);
        
        senderStore.  record (senderRecord);
        receiverStore.record (receiverRecord);
        
        if (!senderPSURL.equals (receiverPSURL)) {
            // Construct exposed interaction metadata
            GlobalPAssertionKey        senderGPAK        = new GlobalPAssertionKey (key, Constants.SENDER_VIEW_TYPE,   lpaid);
            GlobalPAssertionKey        receiverGPAK      = new GlobalPAssertionKey (key, Constants.RECEIVER_VIEW_TYPE, lpaid);
            ExposedInteractionMetadata senderEIM         = new ExposedInteractionMetadata (senderGPAK,   senderMetadata);
            ExposedInteractionMetadata receiverEIM       = new ExposedInteractionMetadata (receiverGPAK, receiverMetadata);
            Record                     senderEIMRecord   = new RecordExposedInteractionMetadata (senderEIM,   key, Constants.SENDER_VIEW_TYPE,   sender.  getElement ());
            Record                     receiverEIMRecord = new RecordExposedInteractionMetadata (receiverEIM, key, Constants.RECEIVER_VIEW_TYPE, receiver.getElement ());
            
            senderStore.  record (senderEIMRecord);
            receiverStore.record (receiverEIMRecord);
        }
        
        return new RecordViews (senderRecord, senderPSURL, receiverRecord, receiverPSURL);
    }
    
    /**
     * Where ObjectIDs are already constructed
     */
    public Record recordRelationship (Record subject, Element subjectAccessor,
            String subjectParameter, List objectIDs, String relation) throws Exception {
        SubjectID              subjectID = new SubjectID (subject.getPAssertion ().getLocalPAssertionID (), subjectAccessor, subjectParameter);
        RelationshipPAssertion relate    = new RelationshipPAssertion (nextLocalPAssertionID (), subjectID, relation, objectIDs);
        InteractionKey         key       = subject.getInteractionKey ();
        Record                 record    = new Record (relate, key, Constants.SENDER_VIEW_TYPE, subject.getAsserter ());
        
        record (record);
        
        return record;
    }
    
    /**
     * One object, no object link
     */
    public Record recordRelationship (Record subject, Element subjectAccessor, String subjectParameter, Record object, Element objectAccessor, String objectParameter, String relation) throws Exception {
        return recordRelationship (subject, subjectAccessor, subjectParameter, object, objectAccessor, objectParameter, null, relation);
    }
    
    /**
     * Records a relationship between two previously recorded p-assertions, one
     * as the subject (with particular data accessor, parameter name) and the
     * other as the object (with particular data accessor, parameter name).
     */
    public Record recordRelationship (Record subject, Element subjectAccessor, String subjectParameter, Record object, Element objectAccessor, String objectParameter, String objectLink, String relation) throws Exception {
        GlobalPAssertionKey gpak = new GlobalPAssertionKey (object.getInteractionKey (), object.getViewKind (), object.getPAssertion ().getLocalPAssertionID ());
        
        return recordRelationship (subject, subjectAccessor, subjectParameter,
                addObjectIDToList (new LinkedList (), gpak, objectParameter, objectAccessor, objectLink),
                relation);
    }
    
    /**
     * Two objects, no object link
     */
    public Record recordRelationship (Record subject, Element subjectAccessor, String subjectParameter, Record object1, Element object1Accessor, String object1Parameter, Record object2, Element object2Accessor, String object2Parameter, String relation) throws Exception {
        return recordRelationship (subject, subjectAccessor, subjectParameter, object1, object1Accessor, object1Parameter, object2, object2Accessor, object2Parameter, null, relation);
    }
    
    /**
     * Records a relationship between three previously recorded p-assertions, one
     * as the subject (with particular data accessor, parameter name) and the
     * other two as objects (with particular data accessors, parameter names).
     */
    public Record recordRelationship (Record subject, Element subjectAccessor,
            String subjectParameter, Record object1, Element object1Accessor,
            String object1Parameter, Record object2, Element object2Accessor,
            String object2Parameter, String objectLink, String relation) throws Exception {
        GlobalPAssertionKey gpak1 = new GlobalPAssertionKey (object1.getInteractionKey (), object1.getViewKind (), object1.getPAssertion ().getLocalPAssertionID ());
        GlobalPAssertionKey gpak2 = new GlobalPAssertionKey (object2.getInteractionKey (), object2.getViewKind (), object2.getPAssertion ().getLocalPAssertionID ());
        
        return recordRelationship (subject, subjectAccessor, subjectParameter,
                addObjectIDToList (
                addObjectIDToList (new LinkedList (), gpak1, object1Parameter, object1Accessor, objectLink),
                gpak2, object2Parameter, object2Accessor, objectLink), relation);
    }
    
    public Record recordRelationship (RecordViews subject, Element subjectAccessor, String subjectParameter,
            RecordViews object1, Element object1Accessor, String object1Parameter, RecordViews object2,
            Element object2Accessor, String object2Parameter, String objectLink, String relation) throws Exception {
        return recordRelationship (subject.getSend (), subjectAccessor, subjectParameter,
                object1.getReceive (), object1Accessor, object1Parameter,
                object2.getReceive (), object2Accessor, object2Parameter, objectLink, relation);
    }
    
    public Record recordRelationship (RecordViews subject, Element subjectAccessor, String subjectParameter,
            RecordViews object1, Element object1Accessor, String object1Parameter, RecordViews object2,
            Element object2Accessor, String object2Parameter, String relation) throws Exception {
        return recordRelationship (subject.getSend (), subjectAccessor, subjectParameter,
                object1.getReceive (), object1Accessor, object1Parameter,
                object2.getReceive (), object2Accessor, object2Parameter, relation);
    }
    
    public Record recordRelationship (RecordViews subject, Element subjectAccessor, String subjectParameter,
            RecordViews object1, Element object1Accessor, String object1Parameter, String relation) throws Exception {
        return recordRelationship (subject.getSend (), subjectAccessor, subjectParameter,
                object1.getReceive (), object1Accessor, object1Parameter, relation);
    }
    
    /**
     * Convenience method for converting serialised XML to a DOM element.
     */
    public Element toDOM (String serialised) throws SAXException, IOException {
        return _domBuilder.parse (new InputSource (new StringReader (serialised))).getDocumentElement ();
    }
    
    /**
     * Performs an XPath query as per the ClientLib Javadoc.
     */
    public abstract Element xpath (String xpath, String root, String item, Map namespaces) throws Exception;
    
    /**
     * Performs an XPath query as per the ClientLib (query method) Javadoc.
     */
    public abstract Element xquery (String xquery) throws Exception;
}
