/********************************************************************************
 * Created on 2006 May 23
 *
 * Copyright (c) 2006 University of Southampton
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *******************************************************************************/
package org.pasoa.preserv.xquery.ipsixq;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;
import org.pasoa.simpledom.*;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;

/**
 * Provides convience methods (write method with different argument types) for
 * writing DOM Documents or other Nodes to String.
 */
public class IPSIXQDOMWriter {
    // Output stream
    private Writer  _out;
    // Should indent output (else no whitespace added)
    private boolean _format;
    private String  _lastXMLNS;
    
    public IPSIXQDOMWriter (Writer output) {
        _out    = new BufferedWriter (output);
        _format = false;
    }
    
    public IPSIXQDOMWriter (OutputStream output) {
        _out    = new OutputStreamWriter (output);
        _format = false;
    }
    
    protected void flush () throws IOException {
        _out.flush ();
    }
    
    protected String getNodeName (Node node) {
        if (node.getNamespaceURI () == null) {
            return node.getNodeName ();
        } else {
            return node.getLocalName ();
        }
    }
    
    protected boolean hasChildren (Element node, boolean onlyElementsCount) {
        NodeList children = node.getChildNodes ();
        boolean  found    = false;
        Node     child;
        
        for (int index = 0; index < children.getLength (); index += 1) {
            child = children.item (index);
            if (child instanceof Element) {
                found = true;
            }
            if (!onlyElementsCount && child instanceof Text) {
                found = true;
            }
        }
        
        return found;
    }
    
    protected void indent (int indent) throws IOException {
        while (indent > 0) {
            write (" ");
            indent -= 1;
        }
    }
    
    protected boolean isXSIType (Attr node) {
        return node.getNamespaceURI () != null && node.getNamespaceURI ().equals (XMLConstants.xsiNamespace) &&
                node.getLocalName ().equals ("type");
    }
    
    public void setFormatted (boolean isFormatted) {
        _format = isFormatted;
    }
    
    protected void write (String text) throws IOException {
        _out.write (text);
        _out.flush ();
    }
    
    public  void write (Attr node) throws IOException {
        write (node, new HashMap ());
    }
    
    private boolean checkPrefixMap (Node node, Map declaredPrefixes) {
        return checkPrefixMap (node.getNodeName (), node.getNodeValue (), declaredPrefixes);
    }
    
    /**
     * Returns true (can be written) if this is not a namespace mapping, or a mapping for a
     * not yet declared prefix.
     * If the given attribute (name, value) is a namespace prefix and not yet
     * declared, add to declaredPrefixes (and return true).
     */
    private boolean checkPrefixMap (String name, String value, Map declaredPrefixes) {
        String prefix;
        
        if (name.startsWith ("xmlns:")) {
            prefix = name.substring (6);
            if (declaredPrefixes.containsKey (prefix)) {
                return false;
            } else {
                declaredPrefixes.put (prefix, value);
            }
        }
        
        return true;
    }
    
    private void write (Attr node, Map declaredPrefixes) throws IOException {
        int colon;
        
        if (node.getNodeName ().endsWith (":xmlns")) {
            return;
        }
        writePrefixMap (node.getPrefix (), declaredPrefixes, node);
        if (isXSIType (node)) {
            colon = node.getNodeValue ().indexOf (":");
            if (colon > 0) {
                writePrefixMap (node.getNodeValue ().substring (0, colon), declaredPrefixes, node);
            }
        }
        if (checkPrefixMap (node, declaredPrefixes)) {
            write (" " + node.getNodeName () + " = \"" + node.getNodeValue () + "\"");
        }
        if (node.getNodeName ().startsWith ("xmlns:") && node.getParentNode () != null &&
                node.getParentNode ().getNodeName ().startsWith (node.getNodeName ().substring (6))) {
            _lastXMLNS = node.getNodeValue ();
        }
        flush ();
    }
    
    /**
     * Write a document out specifying which namespaces have already been declared
     * and the initial indent of elements.
     */
    public void write (Document node, int indent) throws IOException {
        write (node.getDocumentElement (), indent);
        flush ();
    }
    
    /**
     * Convenience method for writing a document with no namespaces already declared
     * and 0 initial indent.
     */
    public void write (Document node) throws IOException {
        write (node, 0);
    }
    
    public void write (Element node) throws IOException {
        write (node, 0);
    }
    
    public void write (Element node, int indent) throws IOException {
        write (node, indent, null, new HashMap ());
    }
    
    /**
     * To change the behaviour of writing out elements, just override this method
     */
    protected void write (Element node, int indent, String lastNamespace, Map declaredPrefixes) throws IOException {
        String qname = getNodeName (node);
        
/*if (qname.endsWith (":Address")) {
            qname += " xmlns = \"" + node.getAttributes ().item (0).getNodeValue () + "\"";
        }*/
        if (_format) {
            indent (indent);
        }
        write ("<" + qname);
        lastNamespace = writeAttributes (node, lastNamespace, declaredPrefixes);
        if (!hasChildren (node, false)) {
            write ("/>");
            if (_format) {
                write ("\n");
            }
        } else {
            write (">");
            if (_format && hasChildren (node, true)) {
                write ("\n");
            }
            writeChildren (node, indent + 1, lastNamespace, declaredPrefixes);
            if (_format && hasChildren (node, true)) {
                indent (indent);
            }
            write ("</" + qname + ">");
            if (_format) {
                write ("\n");
            }
        }
        flush ();
    }
    
    private String substitute (String text, String replace, String replaceWith) {
        int index = -replaceWith.length ();
        
        do {
            index = text.indexOf (replace, index + replaceWith.length ());
            if (index >= 0) {
                text = text.substring (0, index) + replaceWith + text.substring (index + replace.length ());
            }
        } while (index >= 0);
        
        return text;
    }
    
    public void write (Text node) throws IOException {
        String text = node.getNodeValue ();
        
        text = substitute (text, "<", "&lt;");
        text = substitute (text, ">", "&gt;");
        
        write (text);
        flush ();
    }
    
    protected String writeAttributes (Element node, String lastNamespace, Map declaredPrefixes) throws IOException {
        NamedNodeMap children = node.getAttributes ();
        int          colon;
        Node         child;
        
        lastNamespace = writeNamespace (node, lastNamespace);
        for (int index = 0; index < children.getLength (); index += 1) {
            child = children.item (index);
            if (!child.getNodeName ().equals ("xmlns")) {
                write ((Attr) child, declaredPrefixes);
            }
            if (child.getNodeName ().equals ("xmlns") && node.getNamespaceURI () == null) {
                colon = node.getNodeName ().indexOf (":");
                if (colon > 0) {
                    write (" xmlns:" + node.getNodeName ().substring (0, colon) + " = \"" + child.getNodeValue () + "\"");
                }
            }
        }
        return lastNamespace;
    }
    
    protected void writeChildren (Element node, int indent, String lastNamespace, Map declaredPrefixes) throws IOException {
        NodeList children = node.getChildNodes ();
        Node     child;
        
        for (int index = 0; index < children.getLength (); index += 1) {
            child = children.item (index);
            if (child instanceof Element) {
                write ((Element) child, indent, lastNamespace, new HashMap (declaredPrefixes));
            }
            if (child instanceof Text) {
                write ((Text) child);
            }
        }
    }
    
    private String findNamespace (Node node, String prefix, boolean isFirst) {
        NamedNodeMap attrs  = node.getAttributes ();
        Node         parent = node.getParentNode ();
        int          colon;
        
        for (int index = 0; index < attrs.getLength (); index += 1) {
            if (attrs.item (index).getNodeName ().equals ("xmlns:" + prefix) ||
                    attrs.item (index).getNodeName ().equals ("xmlns")) {
                if (!isFirst) {
                    return attrs.item (index).getNodeValue ();
                } else {
                    return null;
                }
            }
        }

        colon = parent.getNodeName ().indexOf (":");
        if (colon > 0) {
            return findNamespace (parent, parent.getNodeName ().substring (0, colon), false);
        } else {
            return null;
        }
    }
    
    protected String writeNamespace (Node node, String lastNamespace) throws IOException {
        int colon;
        String namespace;
        String prefix;
        
        if (node.getNamespaceURI () == null) {
            colon = node.getNodeName ().indexOf (":");
            if (colon > 0) {
                prefix    = node.getNodeName ().substring (0, colon);
                namespace = findNamespace (node, prefix, true);
                if (namespace != null) {
                    write (" xmlns:" + prefix + " = \"" + namespace + "\"");
                }
            } else {
                if (lastNamespace != null && !lastNamespace.equals ("")) {
                    write (" xmlns = \"\"");
                }
                return "";
            }
        }
        if (node.getNamespaceURI () == null || node.getNamespaceURI ().equals (lastNamespace)) {
            return lastNamespace;
        }
        write (" xmlns = \"" + node.getNamespaceURI () + "\"");
        return node.getNamespaceURI ();
    }
    
    private void writePrefixMap (String prefix, Map alreadyDeclared, Node node) throws IOException {
        String space;
        
        if (prefix != null && !prefix.equals ("xmlns")) {
            space = node.lookupNamespaceURI (prefix);
            if (space != null && checkPrefixMap ("xmlns:" + prefix, space, alreadyDeclared)) {
                write (" xmlns:" + prefix + " = \"" + space + "\"");
            }
        }
    }
    
    public static String writeToString (Node toWrite) {
        return writeToString (toWrite, false);
    }
    
    public static String writeToString (Node toWrite, boolean format){
        StringWriter output = new StringWriter ();
        DOMWriter    writer = new DOMWriter (output);
        
        try {
            if (toWrite == null) {
                throw new RuntimeException ("Trying to write null Node");
            }
            writer.setFormatted (format);
            if (toWrite instanceof Attr) {
                writer.write ((Attr) toWrite);
            }
            if (toWrite instanceof Document) {
                writer.write ((Document) toWrite);
            }
            if (toWrite instanceof Element) {
                writer.write ((Element) toWrite);
            }
            if (toWrite instanceof Text) {
                writer.write ((Text) toWrite);
            }
            output.close ();
        } catch (IOException shouldNeverHappen) {
            throw new RuntimeException ("IOException with StringWriter in DOMWriter: " + shouldNeverHappen);
        }
        
        return output.getBuffer ().toString ();
    }
}
