/*Copyright (c) 2005 University of Southampton

 Permission is hereby granted, free of charge,
 to any person obtaining a copy of this
 software and associated documentation files
 (the "Software"), to deal in the Software without
 restriction, including without limitation the rights
 to use, copy, modify, merge, publish, distribute,
 sublicense, and/or sell copies of the Software, and
 to permit persons to whom the Software is furnished
 to do so, subject to the following conditions:

 The above copyright notice and this permission notice
 shall be included in all copies or substantial
 portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY
 OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO
 EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
 FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE
 OR OTHER DEALINGS IN THE SOFTWARE.
 -------------------------------------------------------*/
package org.pasoa.preview;

import java.util.Iterator;
import java.util.Vector;
import org.apache.axis.message.MessageElement;
import org.gridprovenance.client.ProvenanceService;
import org.gridprovenance.client.XQuery;
import org.gridprovenance.client.XQueryResult;
import org.gridprovenance.client.exception.DataConstructorException;
import org.gridprovenance.client.exception.XQueryException;
import org.pasoa.common.BestPractice;
import org.pasoa.common.Constants;
import org.pasoa.pstructure.ObjectID;
import org.pasoa.simpledom.DOMDeserialiser;
import org.pasoa.simpledom.DOMWriter;
import org.w3c.dom.Document;
import org.w3c.dom.Element;


/**
 * Implementation of ModelInterface for PReServ CSL
 * @author tibo
 */
public class ProvenanceRemoteStoreModel implements ModelInterface {
    
    private ProvenanceService provenanceService;
    private AsserterUtility asserterUtility;
    
    /**
     * Create a new instance of ProvenanceRemoteStoreModel
     * @param ps a provenance service
     * @param asserterUtility what kind of asserter do you want to use
     */
    public ProvenanceRemoteStoreModel(ProvenanceService ps, AsserterUtility asserterUtility) {
        this.provenanceService = ps;
        this.asserterUtility = asserterUtility;
    }
    
    /**
     * Get all actors from yours provenance store
     * @return a Vector of AsserterIdentity
     * @throws org.gridprovenance.client.exception.XQueryException CSL Exceptions
     * @throws org.gridprovenance.client.exception.DataConstructorException CSL Exceptions
     */
    public Vector getAsserters() throws XQueryException, DataConstructorException{
        
        XQuery xQueryInstance = provenanceService.newXQuery(asserterUtility.getAssertersQuery());
        XQueryResult xqResult = xQueryInstance.execute();

        Vector tab = new Vector();
        MessageElement[] me = xqResult.getAny();
        if(me != null){   
            Document result = xqResult.getAsDocument();


            DOMDeserialiser domDeserialiser = new DOMDeserialiser(Constants.PSTRUCT_NS);
            Iterator asser = domDeserialiser.iterator(result.getDocumentElement(), Constants.PSTRUCT_NS, "asserter");
            //domDeserialiser.namespace(WSAddressEndpoint.wsaNamespace);
            Element element = null;
            //int i = 0;
           //System.out.println(DOMWriter.writeToString(result,true)); 
           while(asser.hasNext()){
                element =(Element) asser.next();
                if(element != null)
                tab.add(asserterUtility.newAsserterIdentity(element));
            }
            tab.add(asserterUtility.newEmptyAsserterIdentity());
        }
        return tab;
    }
    
    /**
     * Get all interactions between all actors given in argument
     * from yours provenance store
     * @param ActorsNames a Vector of AsserterIdentity
     * @throws org.gridprovenance.client.exception.XQueryException CSL Exceptions
     * @throws org.gridprovenance.client.exception.DataConstructorException CSL Exceptions
     * @throws java.lang.Exception CSL Exceptions
     * @return Vector<Vector<int>> which contains all interactions (like a grid)
     */
    public Vector<Vector> getInteractions(Vector ActorsNames) throws XQueryException, DataConstructorException, Exception{
        
        XQuery xQueryInstance = provenanceService.newXQuery(asserterUtility.getAllInteractions());
        XQueryResult xqResult = xQueryInstance.execute();

        Vector<Vector> tab = new Vector();
        MessageElement[] me = xqResult.getAny();
        if(me != null){   
            //xqResult.printOut();
            Document result = xqResult.getAsDocument();

            for (int i =0 ; i < ActorsNames.size() ; i++){
                Vector tmp = new Vector();
                for (int j =0 ; j < ActorsNames.size() ; j++){
                    tmp.add((int)0);
                }
                tab.add(tmp);
            }


            DOMDeserialiser domDeserialiser = new DOMDeserialiser(Constants.PSTRUCT_NS);
            Iterator ir = domDeserialiser.iterator(result.getDocumentElement() ,Constants.PSTRUCT_NS, "interactionRecord");
            Element elem;
            while(ir.hasNext()){
                elem = (Element)ir.next();
                Element sender = domDeserialiser.single(elem, Constants.PSTRUCT_NS, "sender");
                Element receiver = domDeserialiser.single(elem, Constants.PSTRUCT_NS, "receiver");
                Element asserterS = domDeserialiser.single(sender, Constants.PSTRUCT_NS, "asserter");
                Element asserterR = domDeserialiser.single(receiver, Constants.PSTRUCT_NS, "asserter");

                //System.out.println(DOMWriter.writeToString(sender, true));
                
                int s = asserterUtility.findAsserterIdentity(asserterS, ActorsNames);
                int r = asserterUtility.findAsserterIdentity(asserterR, ActorsNames);
                //System.out.println(s + ", " + r);

                if(s !=-1 && r != -1){
                    Vector tmp = tab.elementAt(s);

                    int nb = Integer.parseInt(tmp.elementAt(r).toString());
                    //System.out.println(nb++);
                    tmp.setElementAt(++nb, r);

                    tab.setElementAt(tmp, s);
                }
                else {
                    throw new Exception("Asserter didn't find!");
                }

            }
        }
        return tab;
    }
    
     
    /**
     * Get all interactions between 2 given actors from yours provenance store
     * @param sender AsserterIdentity as sender
     * @param receiver AsserterIdentity as receiver
     * @throws org.gridprovenance.client.exception.XQueryException CSL Exceptions
     * @throws org.gridprovenance.client.exception.DataConstructorException CSL Exceptions
     * @return Vector which contains all interactions record (XML Element) between these two actors 
     * whith sender as sender and receiver as receiver
     */
    public Vector getAllInteractionsRecordBetween(AsserterIdentity sender, AsserterIdentity receiver) throws XQueryException, DataConstructorException {
       
        XQuery xQueryInstance = provenanceService.newXQuery(asserterUtility.getInteractionsBetweenQuery(sender, receiver));
        XQueryResult xqResult = xQueryInstance.execute();

        //xqResult.printOut();
        
        Vector tab = new Vector();
        MessageElement[] me = xqResult.getAny();
         if(me != null){   
            Document result = xqResult.getAsDocument();

            DOMDeserialiser domDeserialiser = new DOMDeserialiser(Constants.PSTRUCT_NS);
            Iterator inteRecord = domDeserialiser.iterator(result.getDocumentElement(), Constants.PSTRUCT_NS, "interactionRecord");
            Element element;
            Element interkey;
            while(inteRecord.hasNext()){
                interkey = domDeserialiser.single((Element) inteRecord.next(), Constants.PSTRUCT_NS, "interactionKey");
                tab.add(domDeserialiser.text(interkey, "interactionId"));
                //System.out.println(domDeserialiser.text(interkey, "interactionId"));
            }
         }
        return tab;
    }
    
    /**
     * Get interaction record with the given interactionID and between 2 given actors from yours provenance store
     * @param sender AsserterIdentity as sender
     * @param receiver AsserterIdentity as receiver
     * @param interactionID String as interactionID
     * @throws org.gridprovenance.client.exception.XQueryException CSL Exceptions
     * @throws org.gridprovenance.client.exception.DataConstructorException CSL Exceptions
     * @return Element It is a node which contains the searched interaction between these two actors 
     * whith sender as sender and receiver as receiver and interactionID as the id of the interaction
     */
    public Element getInteractionRecordBetween(AsserterIdentity sender, AsserterIdentity receiver, String interactionID) throws XQueryException, DataConstructorException {
       
        XQuery xQueryInstance = provenanceService.newXQuery(asserterUtility.getInteractionRecordBetweenQuery(
                        sender, receiver, interactionID));
        XQueryResult xqResult = xQueryInstance.execute();

        //xqResult.printOut();
        Document result = xqResult.getAsDocument();
      
        DOMDeserialiser domDeserialiser = new DOMDeserialiser(Constants.PSTRUCT_NS);
        Element inteRecord = domDeserialiser.single(result.getDocumentElement(), Constants.PSTRUCT_NS, "interactionRecord");
        //System.out.println(DOMWriter.writeToString(inteRecord, true));
        return inteRecord;
    }
    
     
    /**
     * Get all interactions with the given AsserterIdentity is the receiver from yours provenance store
     * @param ai AsserterIdentity as receiver
     * @throws org.gridprovenance.client.exception.XQueryException CSL Exceptions
     * @throws org.gridprovenance.client.exception.DataConstructorException CSL Exceptions
     * @return Vector which contains all interactions record (XML Element)
     */
    public Vector getAllInteractionReceiverRecord(AsserterIdentity ai) throws XQueryException, DataConstructorException {
       
        XQuery xQueryInstance = provenanceService.newXQuery(asserterUtility.getInteractionsWithReceiverQuery(ai));
        XQueryResult xqResult = xQueryInstance.execute();

        //xqResult.printOut();
        //System.out.println(xqResult);
        
        Document result;
        Vector tab = new Vector();
        MessageElement[] me = xqResult.getAny();
         if(me != null){   
            result = xqResult.getAsDocument();

            DOMDeserialiser domDeserialiser = new DOMDeserialiser(Constants.PSTRUCT_NS);
            Iterator inteRecord = domDeserialiser.iterator(result.getDocumentElement(), Constants.PSTRUCT_NS, "interactionRecord");
            //int i = 0;
            while(inteRecord.hasNext()){
                tab.add((Element) inteRecord.next());
                //System.out.println(DOMWriter.writeToString((Node) tab.elementAt(i), true));
                //i++;
            }
         
         }
         return tab;
    }
     
    /**
     * Get all interactions with the given AsserterIdentity is the sender from yours provenance store
     * @param ai AsserterIdentity as sender
     * @throws org.gridprovenance.client.exception.XQueryException CSL Exceptions
     * @throws org.gridprovenance.client.exception.DataConstructorException CSL Exceptions
     * @return Vector which contains all interactions record (XML Element)
     */
    public Vector getAllInteractionSenderRecord(AsserterIdentity ai) throws XQueryException, DataConstructorException{
       
        XQuery xQueryInstance = provenanceService.newXQuery(asserterUtility.getInteractionsWithSenderQuery(ai));
        XQueryResult xqResult = xQueryInstance.execute();

        //xqResult.printOut();
        //System.out.println(xqResult);
        
        Document result;
        Vector tab = new Vector();
        MessageElement[] me = xqResult.getAny();
         if(me != null){   
            result = xqResult.getAsDocument();

            DOMDeserialiser domDeserialiser = new DOMDeserialiser(Constants.PSTRUCT_NS);
            Iterator inteRecord = domDeserialiser.iterator(result.getDocumentElement(), Constants.PSTRUCT_NS, "interactionRecord");
            while(inteRecord.hasNext()){
                tab.add((Element) inteRecord.next());
                //System.out.println(domDeserialiser.text(interkey, "interactionId"));
            }
        }
        return tab;

    }
    
    /**
     * Get the interactionRecord with the same interactionKey as the given ObjectID from yours provenance store
     * @param ObjectId ObjectID
     * @throws org.gridprovenance.client.exception.DataConstructorException CSL Exceptions
     * @throws org.gridprovenance.client.exception.XQueryException CSL Exceptions
     * @return XML Element which contains an interactionRecord
     */
    public Element getInteractionRecord(ObjectID ObjectId) throws DataConstructorException, XQueryException{
    
        String xQueryString = asserterUtility.getInteractionRecord(ObjectId);
        //System.out.println("Query :");
        //System.out.println(xQueryString);
        XQuery xQueryInstance = provenanceService.newXQuery(xQueryString);
        XQueryResult xqResult = xQueryInstance.execute();
        MessageElement[] me = xqResult.getAny();
        //System.out.println("Result :");
        //xqResult.printOut();
        if(me != null){   
            Document result = xqResult.getAsDocument();
            
            DOMDeserialiser domDeserialiser = new DOMDeserialiser(Constants.PSTRUCT_NS);
            Element e = domDeserialiser.single(result.getDocumentElement(), Constants.PSTRUCT_NS, "interactionRecord");
            return e;
        }
        else return null;
    }
    
    /**
     * Get all the content of the store
     * @throws org.gridprovenance.client.exception.XQueryException CSL Exceptions
     * @throws org.gridprovenance.client.exception.DataConstructorException CSL Exceptions
     * @return XML Document
     */
    public Document getPStruct() throws XQueryException, DataConstructorException{

        String xQueryString ="for $n in $ps:pstruct return $n";
        
        XQuery xQueryInstance = provenanceService.newXQuery(xQueryString);
        XQueryResult xqResult = xQueryInstance.execute();
        
        MessageElement[] me = xqResult.getAny();
        Document result = null;
        if(me != null){   
            //xqResult.printOut();
            result = xqResult.getAsDocument();
        }
        return result;
    }
   
}

