package org.pasoa.pstructure;

import java.io.Serializable;
import java.util.Collections;
import java.util.LinkedList;
import java.util.List;

/**
 * Represents a record of an interaction.
 * @author pgroth, sm
 *
 */
public class InteractionRecord implements Serializable {
    private InteractionKey _interactionKey;
    private List           _senderViews;
    private List           _receiverViews;
    private List           _inferredViews;
    
    /**
     * Construct an InteractionRecord
     *
     * @param recordID the interaction key uniquely identifying this record
     * @param senderView the sender's (client) view of the interaction
     * @param receiverView the recievier's (service) view of the interaction
     */
    public InteractionRecord (InteractionKey interactionKey, List senderViews, List receiverViews) {
        this (interactionKey, senderViews, receiverViews, new LinkedList ());
    }
    
    public InteractionRecord (InteractionKey interactionKey, View senderView, View receiverView) {
        this (interactionKey, makeList (senderView), makeList (receiverView));
    }
    
    /**
     * Used for the q-structure: includes inferred views.
     */
    public InteractionRecord (InteractionKey interactionKey, List senderViews, List receiverViews, List inferredViews) {
        _interactionKey = interactionKey;
        _senderViews    = makeList (senderViews);
        _receiverViews  = makeList (receiverViews);
        _inferredViews  = makeList (inferredViews);
    }
    
    public List getSenderViews () {
        return _senderViews;
    }
    
    /**
     * A client view may return null, if none is present.
     * @return
     */
    public View getSenderView () {
        if (getSenderViews ().size () > 0) {
            return (View) getSenderViews ().get (0);
        } else {
            return null;
        }
    }
    
    public List getInferredViews () {
        return _inferredViews;
    }
    
    public InteractionKey getInteractionKey () {
        return _interactionKey;
    }
    
    /**
     * A service view may return null if none is present
     *
     */
    public View getReceiverView () {
        if (getReceiverViews ().size () > 0) {
            return (View) getReceiverViews ().get (0);
        } else {
            return null;
        }
    }
    
    public List getReceiverViews () {
        return _receiverViews;
    }

    private static List makeList (List original) {
        if (original == null) {
            return new LinkedList ();
        } else {
            return original;
        }
    }
    
    private static List makeList (View item) {
        if (item == null) {
            return new LinkedList ();
        } else {
            return Collections.singletonList (item);
        }
    }
    
    public void setSenderView (View view) {
        _senderViews.add (view);
    }
    
    public void setReceiverView (View view) {
        _receiverViews.add (view);
    }
}
