package org.pasoa.pstructure.deserialiser;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;
import org.pasoa.common.Constants;
import org.pasoa.preserv.xquery.QueryProblem;
import org.pasoa.pstructure.ActorStatePAssertion;
import org.pasoa.pstructure.ExposedInteractionMetadata;
import org.pasoa.pstructure.GlobalPAssertionKey;
import org.pasoa.pstructure.InteractionContext;
import org.pasoa.pstructure.InteractionRecord;
import org.pasoa.pstructure.Link;
import org.pasoa.pstructure.InteractionPAssertion;
import org.pasoa.pstructure.InteractionKey;
import org.pasoa.pstructure.ObjectID;
import org.pasoa.pstructure.PAssertion;
import org.pasoa.pstructure.PAssertionDataKey;
import org.pasoa.pstructure.PHeader;
import org.pasoa.pstructure.RelationshipPAssertion;
import org.pasoa.pstructure.SubjectID;
import org.pasoa.pstructure.View;
import org.pasoa.simpledom.DOMDeserialiser;
import org.pasoa.simpledom.DOMWriter;
import org.w3c.dom.Element;

/**
 * Provides convience methods for deserialising p-structure XML
 * into the classes defined in org.pasoa.common.structure
 * @author sm, pgroth
 *
 */
public class PStructureDeserialiser {
    private DOMDeserialiser _pstruct;
    private DOMDeserialiser _plinks;
    private DOMDeserialiser _precord;
    
    public PStructureDeserialiser () {
        _pstruct = new DOMDeserialiser (Constants.PSTRUCT_NS);
        _plinks  = new DOMDeserialiser (Constants.PLINKS_NS);
        _precord = new DOMDeserialiser (Constants.PRECORD_NS);
    }
    
    public ActorStatePAssertion deserialiseActorStatePAssertion (Element serialised) {
        String localPAssertionID = _pstruct.text (serialised, "localPAssertionId");
        
        return new ActorStatePAssertion (localPAssertionID, serialised);
    }
    
    public ExposedInteractionMetadata deserialiseExposedInteractionMetadata (Element serialised) throws QueryProblem {
        GlobalPAssertionKey globalPAssertionKey = deserialiseGlobalPAssertionKey (_pstruct.single (serialised, "globalPAssertionKey"));
        Iterator            eachMetadata        = _pstruct.iterator (serialised, "interactionMetaData");
        List                interactionMetadata = new LinkedList ();
        
        while (eachMetadata.hasNext ()) {
            interactionMetadata.add (eachMetadata.next ());
        }
        
        return new ExposedInteractionMetadata (globalPAssertionKey, interactionMetadata);
    }
    
    public GlobalPAssertionKey deserialiseGlobalPAssertionKey (Element serialised) {
        InteractionKey  ikey     = deserialiseInteractionKey (_pstruct.single (serialised, "interactionKey"));
        String          viewKind = deserialiseViewKind (_pstruct.single (serialised, "viewKind"));
        String          lpaid    = _pstruct.text (serialised, "localPAssertionId");
        
        return new GlobalPAssertionKey (ikey, viewKind, lpaid);
    }
    
    public InteractionContext deserialiseInteractionContext (Element serialised) throws QueryProblem {
        InteractionKey interactionKey      = deserialiseInteractionKey ((Element) serialised.getElementsByTagNameNS (Constants.PSTRUCT_NS, "interactionKey").item (0).getChildNodes ().item (0));
        List           interactionMetadata = new LinkedList ();
        Iterator       eachMetadata        = _pstruct.iterator (serialised, "interactionMetaData");
        
        while (eachMetadata.hasNext ()) {
            interactionMetadata.add (eachMetadata.next ());
        }
        
        return new InteractionContext (interactionKey, interactionMetadata);
    }
    
    public InteractionKey deserialiseInteractionKey (Element serialised) {
        Element source = _pstruct.single (_pstruct.single (serialised, "messageSource"));
        Element sink   = _pstruct.single (_pstruct.single (serialised, "messageSink"));
        String  iid    = _pstruct.text (serialised, "interactionId");
        
        return new InteractionKey (source, sink, iid);
    }
    
    public InteractionPAssertion deserialiseInteractionPAssertion (Element serialised) {
        String localPAssertionID = _pstruct.text (serialised, "localPAssertionId");
        String transformation    = _pstruct.text (serialised, "documentationStyle");
        
        return new InteractionPAssertion (localPAssertionID, transformation, serialised);
    }
    
    public InteractionRecord deserialiseInteractionRecord (Element serialised) {
        InteractionKey key      = deserialiseInteractionKey (_pstruct.single (serialised, "interactionKey"));
        View           sender   = deserialiseView (_pstruct.single (serialised, "sender"));
        View           receiver = deserialiseView (_pstruct.single (serialised, "receiver"));
        
        return new InteractionRecord (key, sender, receiver);
    }
    
    public Link deserialiseLink (Element serialised) {
        String link = _plinks.text (serialised, "provenanceStoreRef");
        
        if (link == null) {
            return null;
        } else {
            return new Link (link);
        }
    }
    
    public ObjectID deserialiseObjectID (Element serialised) {
        Element             accessor  = _pstruct.single (_pstruct.single (serialised, "dataAccessor"));
        String              parameter = _pstruct.text (serialised, "parameterName");
        GlobalPAssertionKey gpak      = deserialiseGlobalPAssertionKey (serialised);
        Link                link      = deserialiseLink (_plinks.single (serialised, "objectLink"));
        
        return new ObjectID (gpak, parameter, accessor, link);
    }
    
    public PAssertion deserialisePAssertion (Element serialised) throws QueryProblem {
        if (serialised == null) {
            throw new QueryProblem ("No p-assertion to deserialise");
        }
        if (serialised.getLocalName ().equals ("interactionPAssertion")) {
            return deserialiseInteractionPAssertion (serialised);
        }
        if (serialised.getLocalName ().equals ("actorStatePAssertion")) {
            return deserialiseActorStatePAssertion (serialised);
        }
        if (serialised.getLocalName ().equals ("relationshipPAssertion")) {
            return deserialiseRelationshipPAssertion (serialised);
        }
        throw new QueryProblem ("Element does not contain a p-assertion to deserialise: " + DOMWriter.writeToString (serialised));
    }
    
    public PAssertionDataKey deserialisePAssertionDataKey (Element serialised) throws QueryProblem {
        GlobalPAssertionKey gpak     = deserialiseGlobalPAssertionKey (serialised);
        Element             accessor = _pstruct.single (_pstruct.single (serialised, "dataAccessor"));
        
        return new PAssertionDataKey (gpak, accessor);
    }
    
    public PHeader deserialisePHeader (Element serialised) throws QueryProblem {
        InteractionKey  interactionKey      = deserialiseInteractionKey (_pstruct.single (serialised, "interactionKey"));
        Iterator        eachMetadata        = _pstruct.iterator (serialised, "interactionMetaData");
        Iterator        eachContext         = _pstruct.iterator (serialised, "interactionContext");
        List            interactionMetadata = new LinkedList ();
        List            interactionContexts = new LinkedList ();
        
        while (eachMetadata.hasNext ()) {
            interactionMetadata.add (eachMetadata.next ());
        }
        while (eachContext.hasNext ()) {
            interactionContexts.add (deserialiseInteractionContext ((Element) eachContext.next ()));
        }
        
        return new PHeader (interactionKey, interactionMetadata, interactionContexts);
    }
    
    public List deserialisePStructure (Element serialised) {
        List     records    = new LinkedList ();
        Iterator eachRecord = _pstruct.iterator (serialised, "interactionRecord");
        
        while (eachRecord.hasNext ()) {
            records.add (deserialiseInteractionRecord ((Element) eachRecord.next ()));
        }
        
        return records;
    }
    
    public RelationshipPAssertion deserialiseRelationshipPAssertion (Element serialised) {
        String          lpaid         = _pstruct.text (serialised, "localPAssertionId");
        SubjectID       subjectID     = deserialiseSubjectID (_pstruct.single (serialised, "subjectId"));
        String          relation      = _pstruct.text (serialised, "relation");
        Iterator        eachObjectID  = _pstruct.iterator (serialised, "objectId");
        List            objectIDs     = new LinkedList ();
        
        while (eachObjectID.hasNext ()) {
            objectIDs.add (deserialiseObjectID ((Element) eachObjectID.next ()));
        }
        
        return new RelationshipPAssertion (lpaid, subjectID, relation, objectIDs);
    }
    
    public SubjectID deserialiseSubjectID (Element serialised) {
        String  lpaid     = _pstruct.text (serialised, "localPAssertionId");
        Element accessor  = _pstruct.single (_pstruct.single (serialised, "dataAccessor"));
        String  parameter = _pstruct.text (serialised, "parameterName");
        
        return new SubjectID (lpaid, accessor, parameter);
    }
    
    public int deserialiseNumberOfExpectedMsgs (Element serialised) {
        String numStr = serialised.getTextContent ();
        return Integer.parseInt (numStr.trim ());
    }
    
    private View deserialiseView (Element serialised) {
        if (serialised == null) {
            return null;
        }
        
        List     interactionPAssertions     = new LinkedList ();
        List     actorStatePAssertions      = new LinkedList ();
        List     relationshipPAssertions    = new LinkedList ();
        List     exposedInteractionMetadata = new LinkedList ();
        Iterator eachInteraction            = _pstruct.iterator (serialised, "interactionPAssertion");
        Iterator eachActorState             = _pstruct.iterator (serialised, "actorStatePAssertion");
        Iterator eachRelationship           = _pstruct.iterator (serialised, "relationshipPAssertion");
        Element  expectedElement            = _pstruct.single (serialised, "numberOfExpectedAssertions");
        Element  asserter                   = _pstruct.single (_pstruct.single (serialised, "asserter"));
        int      numberOfExpectedPAssertions;
        
        while (eachInteraction.hasNext ()) {
            interactionPAssertions.add (deserialiseInteractionPAssertion ((Element) eachInteraction.next ()));
        }
        while (eachActorState.hasNext ()) {
            actorStatePAssertions.add (deserialiseActorStatePAssertion ((Element) eachActorState.next ()));
        }
        while (eachRelationship.hasNext ()) {
            relationshipPAssertions.add (deserialiseRelationshipPAssertion ((Element) eachRelationship.next ()));
        }
        if (expectedElement != null) {
            numberOfExpectedPAssertions = deserialiseNumberOfExpectedMsgs (expectedElement);
        } else {
            numberOfExpectedPAssertions = -1;
        }
        
        return new View (interactionPAssertions, actorStatePAssertions,
                relationshipPAssertions, exposedInteractionMetadata, numberOfExpectedPAssertions,
                asserter);
    }
    
    public String deserialiseViewKind (Element serialised) {
        String kind;
        
        try {
            kind = _pstruct.stringAttributeNS (serialised, Constants.XSI_NS, "type");
        } catch (NoSuchElementException noAttribute) {
            try {
                kind = _pstruct.stringAttribute (serialised, "xsi:type");
            } catch (NoSuchElementException stupidClientSideLibrary) {
                kind = _pstruct.text (serialised, null, "description");
                if (kind.equals ("isSender"))  {
                    return Constants.SENDER_VIEW_TYPE;
                } else {
                    return Constants.RECEIVER_VIEW_TYPE;
                }
            }
        }
        
        if (kind.endsWith (Constants.SENDER_VIEW_TYPE))  {
            return Constants.SENDER_VIEW_TYPE;
        } else {
            return Constants.RECEIVER_VIEW_TYPE;
        }
    }
}
