package org.pasoa.storeclient.pquery;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * Builder for a query data handle.  Use this class to specify all the criteria
 * by which the data item of which you want to find the provenance is determinable.
 */
public class QDHBuilder {
    private String  _source;
    private String  _sink;
    private boolean _sender;
    private boolean _receiver;
    private boolean _isInteraction;
    private List    _contents;
    private List    _concurrents;
    private List    _concurViews;
    private String  _itemPath;
    
    public QDHBuilder () {
        _source        = null;
        _sink          = null;
        _sender        = true;
        _receiver      = true;
        _isInteraction = true;
        _contents      = new LinkedList ();
        _concurrents   = new LinkedList ();
        _concurViews   = new LinkedList ();
        _itemPath      = null;
    }
    
    /**
     * Returns an XPath string for the query data handle.  Used by ProvenanceQueryBuilder,
     * other clients should not normally need to use this method.
     */
    public String build () {
        StringBuffer path        = new StringBuffer ("/ps:pstruct/ps:interactionRecord");
        Iterator     eachConcur  = _concurrents.iterator ();
        Iterator     eachContent = _contents.   iterator ();
        Iterator     eachConView = _concurViews.iterator ();
        
        if (_source != null) {
            path.append ("[ps:interactionKey/ps:messageSource/wsa:Address='" + _source + "']");
        }
        if (_sink != null) {
            path.append ("[ps:interactionKey/ps:messageSink/wsa:Address='" + _sink + "']");
        }
        // Checks in other view
        if (eachConView.hasNext ()) {
            if (_sender) {
                if (_receiver) {
                    path.append ("/*");
                } else {
                    path.append ("/ps:receiver");
                }
            } else {
                path.append ("/ps:sender");
            }
            while (eachConView.hasNext ()) {
                path.append (eachConView.next ());
            }
            path.append ("/..");
        }
        // Check in this view
        if (_sender) {
            if (_receiver) {
                path.append ("/*");
            } else {
                path.append ("/ps:sender");
            }
        } else {
            path.append ("/ps:receiver");
        }
        while (eachConcur.hasNext ()) {
            path.append (eachConcur.next ());
        }
        if (_isInteraction) {
            path.append ("/ps:interactionPAssertion");
        } else {
            path.append ("/ps:actorStatePAssertion");
        }
        while (eachContent.hasNext ()) {
            path.append ("/ps:content" + eachContent.next () + "/..");
        }
        if (_itemPath != null) {
            path.append ("/ps:content" + _itemPath);
        }
        
        return path.toString ();
    }
    
    /**
     * Specify that the message source of the interaction containing the data item has a given
     * WS-Addressing Endpoint address URL.
     */
    public QDHBuilder sourceURL (String source) {
        _source = source;
        return this;
    }
    
    /**
     * Specify that the message sink of the interaction containing the data item has a given
     * WS-Addressing Endpoint address URL.
     */
    public QDHBuilder sinkURL (String sink) {
        _sink = sink;
        return this;
    }
    
    /**
     * Specify that the data item was asserted in the sender view (without calling
     * this or receiver() both views will be searched).
     */
    public QDHBuilder sender () {
        _receiver = false;
        return this;
    }
    
    /**
     * Specify that the data item was asserted in the receiver view (without calling
     * this or sender() both views will be searched).
     */
    public QDHBuilder receiver () {
        _sender = false;
        return this;
    }
    
    /**
     * Specify that the data item is, or is part of, an interaction p-assertion
     * (this is the default, method included for completeness).
     */
    public QDHBuilder interactionItem () {
        _isInteraction = true;
        return this;
    }
    
    /**
     * Specify that the data item is, or is part of, an actor state p-assertion.
     */
    public QDHBuilder stateItem () {
        _isInteraction = false;
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * contents of the p-assertion containing the data item.
     */
    public QDHBuilder contentCheck (String xpath) {
        _contents.add ("[" + xpath + "]");
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * contents of the SOAP Body in the p-assertion containing the data item.
     */
    public QDHBuilder soapBodyCheck (String xpath) {
        _contents.add ("/soap:Envelope/soap:Body[" + xpath + "]/../..");
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * contents of an actor state p-assertion in the same view as the data item.
     */
    public QDHBuilder concurrentState (String xpath) {
        _concurrents.add ("/ps:actorStatePAssertion/ps:content[" + xpath + "]/../..");
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * contents of an interaction p-assertion in the same view as the data item.
     */
    public QDHBuilder concurrentInteraction (String xpath) {
        _concurrents.add ("/ps:interactionPAssertion/ps:content[" + xpath + "]../..");
        return this;
    }
    
    /**
     * Specify the path into the contents of the p-assertion to find the data
     * item of interest (i.e. a data accessor).
     */
    public QDHBuilder contentItem (String xpath) {
        _itemPath = xpath;
        return this;
    }
    
    /**
     * Specify the path into the SOAP Body contents of the p-assertion to find the data
     * item of interest (i.e. a data accessor).
     */
    public QDHBuilder soapBodyItem (String xpath) {
        _itemPath = "/soap:Envelope/soap:Body" + xpath;
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * contents of an actor state p-assertion in the other view from the data item.
     */
    public QDHBuilder concurrentViewState (String xpath) {
        _concurViews.add ("/ps:actorStatePAssertion/ps:content[" + xpath + "]/../..");
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * contents of an interaction p-assertion in the other view from the data item.
     */
    public QDHBuilder concurrentViewInteraction (String xpath) {
        _concurViews.add ("/ps:interactionPAssertion/ps:content[" + xpath + "]../..");
        return this;
    }
}
