package org.pasoa.storeclient.pquery;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * Builder for a relationship target filter.  Use this class to specify all the criteria
 * by which a data item is determined to be in scope.
 *
 * Currently, this class builds 'excluding' paths, i.e. everything specified using
 * this restricts (rather than expands) the scope.
 */
public class RTFBuilder {
    private boolean _excluding;
    private String  _relation;
    private List    _notRelations;
    private List    _notParameters;
    private String  _parameter;
    private String  _source;
    private String  _sink;
    private boolean _isInteraction;
    private boolean _isState;
    private List    _contents;
    private List    _concurrents;
    
    public RTFBuilder () {
        _excluding     = true;
        _relation      = null;
        _parameter     = null;
        _source        = null;
        _sink          = null;
        _isInteraction = false;
        _isState       = false;
        _contents      = new LinkedList ();
        _concurrents   = new LinkedList ();
        _notRelations  = new LinkedList ();
        _notParameters = new LinkedList ();
    }
    
    /**
     * Returns an XPath string for the relationship target filter.  Used by ProvenanceQueryBuilder,
     * other clients should not normally need to use this method.
     */
    public String build () {
        StringBuffer path         = new StringBuffer ("/pq:relationshipTarget");
        Iterator     eachContent  = _contents.iterator ();
        Iterator     eachConcur   = _concurrents.iterator ();
        Iterator     eachNotRel   = _notRelations.iterator ();
        Iterator     eachNotParam = _notParameters.iterator ();
        
        if (_relation != null) {
            path.append ("[ps:relation='" + _relation + "']");
        }
        while (eachNotRel.hasNext ()) {
            path.append ("[fn:not(ps:relation='" + eachNotRel.next () + "')]");
        }
        if (_parameter != null) {
            path.append ("[ps:parameterName='" + _parameter + "']");
        }
        while (eachNotParam.hasNext ()) {
            path.append ("[fn:not(ps:parameterName='" + eachNotParam.next () + "')]");
        }
        if (_source != null) {
            path.append ("[ps:interactionKey/ps:messageSource/wsaAddress='" + _source + "']");
        }
        if (_sink != null) {
            path.append ("[ps:interactionKey/ps:messageSink/wsaAddress='" + _sink + "']");
        }
        while (eachConcur.hasNext ()) {
            path.append ("[ps:interactionRecord/*" + eachConcur.next ());
        }
        if (_isInteraction) {
            path.append ("/ps:interactionPAssertion");
        }
        if (_isState) {
            path.append ("/ps:actorStatePAssertion");
        }
        if (_isInteraction || _isState) {
            while (eachContent.hasNext ()) {
                path.append ("/ps:content" + eachContent.next () + "/..");
            }
        }
        
        return path.toString ();
    }
    
    /**
     * Specifies the relation that the target must have.
     */
    public RTFBuilder relation (String relation) {
        _relation = relation;
        return this;
    }
    
    /**
     * Specifies a relation that the target must not have.
     */
    public RTFBuilder notRelation (String relation) {
        _notRelations.add (relation);
        return this;
    }
    
    /**
     * Specifies the parameter name that the target must have.
     */
    public RTFBuilder parameter (String parameter) {
        _parameter = parameter;
        return this;
    }
    
    /**
     * Specifies a parameter name that the target must not have.
     */
    public RTFBuilder notParameter (String parameter) {
        _notParameters.add (parameter);
        return this;
    }
    
    /**
     * Specifies that the message source of the interaction containing the
     * target must have a given WS-Addressing Endpoint address URL.
     */
    public RTFBuilder sourceURL (String source) {
        _source = source;
        return this;
    }
    
    /**
     * Specifies that the message sink of the interaction containing the
     * target must have a given WS-Addressing Endpoint address URL.
     */
    public RTFBuilder sinkURL (String sink) {
        _sink = sink;
        return this;
    }
    
    /**
     * Specifies that the target must be (in) an interaction p-assertion
     * (default).
     */
    public RTFBuilder isInteraction () {
        _isInteraction = true;
        _isState       = false;
        return this;
    }
    
    /**
     * Specifies that the target must be (in) an actor state p-assertion.
     */
    public RTFBuilder isState () {
        _isInteraction = false;
        _isState       = true;
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * contents of the p-assertion containing the data item.
     */
    public RTFBuilder contentCheck (String xpath) {
        _contents.add ("[" + xpath + "]");
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * SOAP Body contents of the p-assertion containing the data item.
     */
    public RTFBuilder soapBodyCheck (String xpath) {
        _contents.add ("/soap:Envelope/soap:Body[" + xpath + "]/../..");
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * contents of an actor state p-assertion in the interaction record of the target.
     */
    public RTFBuilder concurrentState (String xpath) {
        _concurrents.add ("/ps:actorStatePAssertion/ps:content[" + xpath + "]");
        return this;
    }
    
    /**
     * Specify an XPath predicate check (contents of a [...]) on the
     * contents of an interaction p-assertion in the interaction record of the target.
     */
    public RTFBuilder concurrentInteraction (String xpath) {
        _concurrents.add ("/ps:interactionPAssertion/ps:content[" + xpath + "]");
        return this;
    }
}
