package org.pasoa.storeclient.wrappers.java;

import java.io.File;
import java.io.IOException;
import java.io.StringReader;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import org.pasoa.preserv.xquery.QueryProblem;
import org.pasoa.pstructure.InteractionKey;
import org.pasoa.pstructure.PHeader;
import org.pasoa.pstructure.deserialiser.PStructureDeserialiser;
import org.pasoa.applications.pce.docstyles.SequentialArgumentsRequest;
import org.pasoa.simpledom.SimpleDocument;
import org.pasoa.simpledom.SimpleElement;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Wraps a Java class' main method as a service.  This class also provides
 * general functionality for any service invoked with a sequence of arguments,
 * as in a command-line.  The run method can be overridden to cause other
 * services of this form to be executed.
 */
public class JavaService implements InvokedActor {
    private String                     _name;
    private List                       _parameters;
    private List                       _inputs;
    private List                       _outputs;
    private String                     _identity;
    private String                     _requestInteraction;
    private String                     _responseInteraction;
    private String                     _pHeaderFile;
    private SequentialArgumentsRequest _request;
    
    public JavaService (String asserterIdentity, String scriptName) {
        _identity            = asserterIdentity;
        _name                = scriptName;
        _parameters          = new LinkedList ();
        _inputs              = new LinkedList ();
        _outputs             = new LinkedList ();
        _request             = new SequentialArgumentsRequest ();
        _requestInteraction  = UUID.randomUUID ().toString ();
        _responseInteraction = UUID.randomUUID ().toString ();
        
        _request.setName (_name);
    }
    
    public JavaService (Map config) {
        _identity   = ((List) config.get ("identity")).get (0).toString ();
        _name       = ((List) config.get ("command")).get (0).toString ();
        _parameters = (List)  config.get ("parameters");
        _inputs     = (List)  config.get ("inputs");
        _outputs    = (List)  config.get ("outputs");
        _request             = new SequentialArgumentsRequest ();
        _requestInteraction  = UUID.randomUUID ().toString ();
        _responseInteraction = UUID.randomUUID ().toString ();
        
        _request.setName (_name);
    }
    
    protected void addArgument (String parameter, String value) {
        _request.addArgument (parameter, value);
    }
    
    public Element getArgumentDataAccessor (Object argument) {
        try {
            return toElement ("<xpath>//sads:argument[" + _request.getArguments ().indexOf (argument) + "]/value</xpath>");
        } catch (ParserConfigurationException ex) {
            throw new RuntimeException (ex);
        } catch (SAXException ex) {
            throw new RuntimeException (ex);
        } catch (IOException ex) {
            throw new RuntimeException (ex);
        }
    }
    
    protected void addArguments (String parameter, String[] values) {
        for (String value : values) {
            addArgument (parameter, value);
        }
    }
    
    protected void addInput (String newInput) {
        _inputs.add (newInput);
    }
    
    protected void addOutput (String newOutput) {
        _outputs.add (newOutput);
    }
    
    public Element getIdentity () {
        return getRequestID ().getMessageSink ();
    }
    
    public List getInputs () {
        List results = new LinkedList ();
        int  numArgs = _request.getArguments ().size ();
        
        for (int index = 0; index < numArgs; index += 1) {
            if (_inputs.contains (_request.getParameter (index))) {
                results.add (_request.getArgument (index));
            }
        }
        
        return results;
    }
    
    public String getName () {
        return _name;
    }
    
    public List getOutputs () {
        List results = new LinkedList ();
        int  numArgs = _request.getArguments ().size ();
        
        for (int index = 0; index < numArgs; index += 1) {
            if (_outputs.contains (_request.getParameter (index))) {
                results.add (_request.getArgument (index));
            }
        }
        
        return results;
    }
    
    public String getPHeaderFile () {
        return _pHeaderFile;
    }
    
    public Element getRequest () {
        return _request.getStyledMessage ().getDocumentElement ();
    }
    
    public InteractionKey getRequestID () {
        Document doc    = new SimpleDocument ();
        Element  condor = new SimpleElement  (doc, "", "condor");
        Element  me     = new SimpleElement  (doc, "", _identity);
        
        return new InteractionKey (condor, me, _requestInteraction);
    }
    
    public Element getResponse () {
        SimpleElement ack = new SimpleElement  (new SimpleDocument (), "", "acknowledgement");
        
        ack.appendChild (_request.getPHeader ());
        
        return ack;
    }
    
    public InteractionKey getResponseID () {
        Document doc    = new SimpleDocument ();
        Element  condor = new SimpleElement  (doc, "", "condor");
        Element  me     = new SimpleElement  (doc, "", _identity);
        
        return new InteractionKey (me, condor, _responseInteraction);
    }
    
    public void initialise (String[] confargs) {
        List argList = Arrays.asList (confargs);
        
        try {
            setPHeader ((String) argList.get (0));
            _request.addArguments (_parameters, argList.subList (1, argList.size ()));
        } catch (QueryProblem ex) {
            throw new RuntimeException (ex);
        }
    }
    
    public void run () {
        try {
            String[] args = _request.getArgumentsStringArray ();
            Class.forName (_name).getMethod ("main", new Class [] {String[].class}).invoke (null, new Object[] {args});
        } catch (Exception problem) {
            throw new RuntimeException (problem);
        }
    }
    
    public void setPHeader (String pheaderFile) throws QueryProblem {
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance ();
            factory.setNamespaceAware (true);
            Element serialised = factory.newDocumentBuilder ().parse (new File (pheaderFile)).getDocumentElement ();
            PHeader template   = new PStructureDeserialiser ().deserialisePHeader (serialised);
            PHeader pheader    = new PHeader (getRequestID (), template.getInteractionMetadata (), template.getInteractionContexts ());

            _pHeaderFile = pheaderFile;
            _request.setPHeader (pheader);
        } catch (SAXException ex) {
            ex.printStackTrace ();
        } catch (IOException ex) {
            ex.printStackTrace ();
        } catch (ParserConfigurationException ex) {
            ex.printStackTrace ();
        }
    }
    
    public static Element toElement (String value) throws IOException, ParserConfigurationException, SAXException {
        return DocumentBuilderFactory.newInstance ().newDocumentBuilder ().parse (new InputSource (new StringReader (value))).getDocumentElement ();
    }
    
    protected String toString (Object value) {
        String vname = "";
        
        if (value instanceof File) {
            return ((File) value).getName ();
        }
        if (value instanceof Map) {
            for (Character group : ((Map <Character, String>) value).keySet ()) {
                vname = vname + group + ":" + ((Map <Character, String>) value).get (group) + ";";
            }
            return vname;
        }
        return value.toString ();
    }
}
